"""Top-level package for GisMap."""

from importlib.metadata import metadata

from gismap.sources.hal import HAL as HAL, HALAuthor as HALAuthor
from gismap.sources.dblp import DBLP as DBLP, DBLPAuthor as DBLPAuthor
from gismap.sources.ldb import LDB as LDB, LDBAuthor as LDBAuthor
from gismap.utils.common import get_classes as get_classes
from gismap.gismo import make_gismo as make_gismo
from gismap.search import (
    Search as Search,
    SearchDocuments as SearchDocuments,
    SearchLandmarks as SearchLandmarks,
    SearchFeatures as SearchFeatures,
    search_to_html as search_to_html,
    search_to_text as search_to_text,
)
from gismap.gisgraphs.widget import GismapWidget as GismapWidget


infos = metadata(__name__)
__version__ = infos["Version"]
__author__ = """Fabien Mathieu"""
__email__ = "fabien.mathieu@normalesup.org"



import re
from bs4 import BeautifulSoup as Soup
from gismap.lab.labmap import LabMap
from gismap.lab.lab_author import AuthorMetadata, LabAuthor
from gismap.utils.requests import get


def name_changer(name, rosetta):
    return rosetta.get(name, name)


class LaasMap(LabMap):
    """
    Class for handling a LAAS team from its name.
    Default to `sara` team.
    """

    name = "sara"
    base_url = "https://www.laas.fr"
    rosetta = {"Urtzi Ayesta Morate": "Urtzi Ayesta"}

    def _author_iterator(self):
        soup = Soup(get(f"{self.base_url}/fr/equipes/{self.name}/"), features="lxml")
        for a in soup("div", {"class": "membre"})[0]("a"):
            url = self.base_url + a["href"]
            name = name_changer(a.img["alt"], self.rosetta)
            img = (
                self.base_url + a.img["src"]
                if "public_avatar" in a.img["class"]
                else None
            )
            yield LabAuthor(
                name=name,
                metadata=AuthorMetadata(url=url, img=img, group=self.name.upper()),
            )


class LaasFull(LabMap):
    """
    Class for handling all LAAS teams using `https://www.laas.fr/fr/equipes/` to get team names.
    """

    name = "LAAS"

    def _author_iterator(self):
        soup = Soup(get("https://www.laas.fr/fr/equipes/"), features="lxml")
        teams = [a["href"].split("/")[-2] for a in soup("a", {"class": "badge"})]
        for team in teams:
            for author in LaasMap(name=team)._author_iterator():
                yield author


class SolaceMap(LabMap):
    """
    Class for handling the Solace team (`https://solace.cnrs.fr`).
    """

    name = "Solace"
    regex = re.compile(r"<li>(.*?)(,| \(|</li>)")

    def _author_iterator(self):
        html = get("https://solace.cnrs.fr/people.html", verify=False)
        for name, _ in self.regex.findall(html):
            soup = Soup(name, features="lxml")
            url = soup.a["href"] if soup.a else None
            yield LabAuthor(
                name=soup.text.strip(),
                metadata=AuthorMetadata(url=url, group=self.name.upper()),
            )



from collections import defaultdict
from string import Template

from gismap.utils.text import reduce_keywords, Corrector


class SearchAction:
    """
    Base class for extracting search results from a Gismo.

    Subclasses should implement :meth:`process` to define
    how to extract results from the gismo.

    Parameters
    ----------
    name : :class:`str`, optional
        Name of this action (used as key in results dict).
    post : callable, optional
        Post-processing function applied to results.
    """

    def __init__(self, name=None, post=None):
        self.name = name
        self.post = (lambda x: x) if post is None else post

    def process(self, gismo):
        """
        Extract results from the gismo. Must be implemented by subclasses.

        Parameters
        ----------
        gismo : :class:`~gismo.gismo.Gismo`
            The gismo to query.

        Returns
        -------
        Results (type depends on subclass).
        """
        raise NotImplementedError

    def run(self, gismo):
        """
        Execute the action and apply post-processing.

        Parameters
        ----------
        gismo : :class:`~gismo.gismo.Gismo`
            The gismo to query.

        Returns
        -------
        Post-processed results.
        """
        return self.post(self.process(gismo))


def p2t(publis):
    """
    Parameters
    ----------
    publis: :class:`list`
        List of publications

    Returns
    -------
    :class:`str`
        Publications converted in text and concatenated.
    """
    return "\n".join(a.string for a in publis)


def l2t(lis):
    """

    Parameters
    ----------
    lis: :class:`list`
        List of text.
    Returns
    -------
    :class:`str`
        Concatenation, comma-separated.
    """
    return ", ".join(lis)


class SearchDocuments(SearchAction):
    """Gives *k* best covering articles."""

    def __init__(self, name="articles", post=None, k=5):
        if post is None:
            post = p2t
        super().__init__(name=name, post=post)
        self.k = k

    def process(self, gismo):
        return gismo.get_documents_by_coverage(k=self.k)


class SearchFeatures(SearchAction):
    """Gives best keywords."""

    def __init__(self, name="keywords", post=None):
        if post is None:
            post = l2t
        super().__init__(name=name, post=post)

    def process(self, gismo):
        return reduce_keywords(gismo.get_features_by_rank())


class SearchLandmarks(SearchAction):
    """Gives best landmarks."""

    def __init__(self, name="landmarks", post=None, lmks=None):
        if post is None:
            post = l2t
        super().__init__(name=name, post=post)
        self.lmks = lmks

    def process(self, gismo):
        return self.lmks.get_landmarks_by_rank(gismo)


class Search:
    """
    Builds a gismo search engine.

    Parameters
    ----------
    gismo: :class:`~gismo.gismo.Gismo`
        Gismo to use.
    action_list: :class:`list`
        List of actions to perform.
    post: callable, optional
        Output transformation.
    corrector: :class:`Bool`, default=True
        Implement word correction.
    """

    def __init__(self, gismo, action_list, post=None, corrector=True):
        self.gismo = gismo
        self.action_list = action_list
        self.post = (lambda x: x) if post is None else post
        if corrector:
            self.corrector = Corrector(gismo.embedding.features)
        else:
            self.corrector = None

    def __call__(self, query):
        if self.corrector is not None:
            query = self.corrector(query)
        success = self.gismo.rank(query)
        res = dict()
        if success:
            for action in self.action_list:
                res[action.name] = action.run(self.gismo)
        return self.post({"query": query, "success": success, "results": res})


def search_to_text(res):
    """
    Parameters
    ----------
    res: :class:`dict`
        Raw results of search.

    Returns
    -------
    :class:`str`
        Text representation of the results.
    """
    query = res["query"]
    if not res["success"]:
        return f"Failure: ``{query}'' not found!"
    output = f"Results for ``{query}'':\n"
    for k, v in res["results"].items():
        output += f"Suggested {k}: {v}\n"
    return output


publi_template = Template("""
<li>
<i>$title</i>, by $authors. $venue, $year. $hal $dblp
</li>
""")


def publi_to_html(publi):
    """
    Convert a publication to an HTML list item.

    Parameters
    ----------
    publi : :class:`~gismap.sources.models.Publication`
        Publication to convert.

    Returns
    -------
    :class:`str`
        HTML list item string.
    """
    dico = dict()
    for db in ["hal", "dblp"]:
        source = publi.sources.get(db)
        if source:
            dico[db] = f"<a href='{source['url']}' target='_blank'>{db.upper()}</a>"
        else:
            dico[db] = ""
    dico["authors"] = ", ".join(a.name for a in publi.authors)
    for key in ["title", "venue", "year"]:
        dico[key] = getattr(publi, key)
    return publi_template.substitute(dico)


def publis_to_html(publis):
    """
    Convert a list of publications to an HTML unordered list.

    Parameters
    ----------
    publis : :class:`list`
        List of publications.

    Returns
    -------
    :class:`str`
        HTML unordered list string.
    """
    rows = "\n".join(publi_to_html(p) for p in publis)
    return f"<ul>\n{rows}\n</ul>"


html_template = Template("""
<div>
<h4>Search: <i>$query</i></h4>
<div>
<h5>Associated keywords:</h5>
<div>$keywords</div>
</div>
<div>
<h5>Associated Projects:</h5>
<div>$projects</div>
</div>
<div>
<h5>Suggested people:</h5>
<div>$members</div>
</div>
<div>
<h5>Suggested publications:</h5>
<div>$publis</div>
</div>
</div>

""")


def search_to_html(res):
    """
    Parameters
    ----------
    res: :class:`dict`
        Raw results of search.

    Returns
    -------
    :class:`str`
        HTML representation of the results.
    """
    dico = defaultdict(str)
    dico.update(res["results"])
    dico["query"] = res["query"]
    if res["success"]:
        dico["publis"] = publis_to_html(dico["articles"])
    return html_template.safe_substitute(dico)



from dataclasses import dataclass
from typing import ClassVar

from gismap.utils.common import LazyRepr


@dataclass(repr=False)
class Author(LazyRepr):
    """
    Base class for authors in the database system.

    Authors are identified primarily by their name and may have
    database-specific subclasses with additional attributes like keys and aliases.

    Parameters
    ----------
    name : :class:`str`
        The author's name.
    """

    name: str


@dataclass(repr=False)
class Publication(LazyRepr):
    """
    Base class for publications in the database system.

    Publications contain metadata about academic papers including title,
    authors, venue, type, and publication year.

    Parameters
    ----------
    title : :class:`str`
        The publication title.
    authors : :class:`list`
        List of :class:`Author` objects.
    venue : :class:`str`
        Publication venue (journal, conference, etc.).
    type : :class:`str`
        Publication type (e.g., 'journal', 'conference', 'book').
    year : :class:`int`
        Year of publication.
    """

    title: str
    authors: list
    venue: str
    type: str
    year: int


@dataclass(repr=False)
class DB(LazyRepr):
    """
    Abstract base class for database backends.

    Provides the interface for searching authors and retrieving publications.
    Subclasses must implement :meth:`search_author` and :meth:`from_author`.

    Attributes
    ----------
    db_name : :class:`str`
        Identifier for the database backend (e.g., 'hal', 'dblp', 'ldb').
    """

    db_name: ClassVar[str] = None

    @classmethod
    def search_author(cls, name):
        """
        Search for authors matching the given name.

        Parameters
        ----------
        name : :class:`str`
            Name to search for.

        Returns
        -------
        :class:`list`
            List of matching :class:`Author` objects.
        """
        raise NotImplementedError

    @classmethod
    def from_author(cls, a):
        """
        Retrieve publications for a given author.

        Parameters
        ----------
        a : :class:`Author`
            The author whose publications to retrieve.

        Returns
        -------
        :class:`list`
            List of :class:`Publication` objects.
        """
        raise NotImplementedError


def db_class_to_auth_class(db_class):
    """
    Find the Author subclass associated with a given DB class.

    Parameters
    ----------
    db_class : :class:`type`
        A DB subclass (e.g., HAL, DBLP, LDB).

    Returns
    -------
    :class:`type` or None
        The corresponding Author subclass, or None if not found.
    """
    for subclass in Author.__subclasses__():
        if db_class in subclass.__mro__:
            return subclass
    return None



from time import sleep
import requests
from importlib.metadata import metadata
from gismap.utils.logger import logger


infos = metadata("gismap")
session = requests.Session()
session.headers.update(
    {
        "User-Agent": f"{infos['name']}/{infos['Version']} ({'; '.join(infos.get_all('Project-URL'))}; Contact, {infos['author-email']}"
    }
)


def get(url, params=None, n_trials=10, verify=True, encoding=None):
    """
    Parameters
    ----------
    url: :class:`str`
        Entry point to fetch.
    params: :class:`dict`, optional
        Get arguments (appended to URL).
    n_trials: :class:`int`, default=10
        Number of attempts to fetch URL.
    verify: :class:`bool`, default=True
        Verify certificates.
    encoding: :class:`str`, optional
        Force response encoding (e.g. ``"utf-8"``). Useful when the server
        does not declare the charset and ``requests`` falls back to ISO-8859-1.

    Returns
    -------
    :class:`str`
        Result.
    """
    for attempt in range(n_trials):
        try:
            r = session.get(url, params=params, verify=verify)
            if r.status_code == 429:
                try:
                    t = int(r.headers["Retry-After"])
                except KeyError:
                    t = 60
                logger.warning(f"Too many requests. Auto-retry in {t} seconds.")
                sleep(t)
            else:
                if encoding is not None:
                    r.encoding = encoding
                return r.text
        except requests.exceptions.ConnectionError:
            t = 6
            logger.warning(f"Connection error. Auto-retry in {t} seconds.")
            sleep(t)
    raise requests.exceptions.ConnectionError(f"Unable to retrieve {url}")



from gismo.common import MixInIO
import zstandard as zstd
import numpy as np
import pickle

dctx = zstd.ZstdDecompressor()
cctx = zstd.ZstdCompressor()


class ZList(MixInIO):
    """
    Compressed list with frame-based storage.

    Stores elements in compressed frames, allowing efficient memory usage
    while maintaining random access. Uses zstandard compression.

    Use as a context manager for building:

        with ZList(frame_size=1000) as z:
            for item in data:
                z.append(item)

    Parameters
    ----------
    frame_size : :class:`int`, default=1000
        Number of elements per compressed frame.
    """

    def __init__(self, frame_size=1000):
        self.frame_size = frame_size
        self.frame = None
        self._frame_index = None
        self._blob = None
        self._off = None
        self._n = None
        self._batch = None

    def _merge_batch(self):
        if self._batch:
            frame = cctx.compress(pickle.dumps(self._batch))
            self._blob += frame
            self._off.append(len(self._blob))
            self._batch = []

    def append(self, entry):
        """
        Add an element to the list.

        Parameters
        ----------
        entry
            Element to add.
        """
        self._batch.append(entry)
        self._n += 1
        if len(self._batch) == self.frame_size:
            self._merge_batch()

    @property
    def size(self):
        return len(self._blob)

    def __enter__(self):
        self._blob = bytearray()
        self._off = [0]
        self._n = 0
        self._batch = []
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._merge_batch()
        self._blob = bytes(self._blob)
        self._off = np.array(self._off, dtype=int)

    def load_frame(self, f):
        self.frame = pickle.loads(
            dctx.decompress(self._blob[self._off[f] : self._off[f + 1]])
        )

    def __getitem__(self, i):
        g, f = i // self.frame_size, i % self.frame_size
        if g != self._frame_index:
            self.load_frame(g)
            self._frame_index = g
        return self.frame[f]

    def __len__(self):
        return self._n



from dataclasses import dataclass, field
from bof.fuzz import jit_square_factors
from bof.feature_extraction import CountVectorizer
import numpy as np
import unicodedata

from gismap.sources.models import Publication, Author
from gismap.utils.text import clean_aliases


def score_author_source(dbauthor):
    """
    Compute a quality score for an author source.

    Higher scores indicate more reliable author identification.
    HAL idHal keys are preferred, followed by HAL pid, then DBLP/LDB.

    Parameters
    ----------
    dbauthor: :class:`~gismap.sources.models.Author`
        A database-specific author object.

    Returns
    -------
    :class:`int`
        Score value (higher is better).
    """
    if dbauthor.db_name == "hal":
        if dbauthor.key_type == "fullname":
            return -1
        elif dbauthor.key_type == "pid":
            return 2
        else:
            return 3
    elif dbauthor.db_name in ["dblp", "ldb"]:
        return 1
    else:
        return 0


def sort_author_sources(sources):
    """
    Sort author sources by quality score in descending order.

    Parameters
    ----------
    sources: :class:`list`
        List of database-specific author objects.

    Returns
    -------
    :class:`list`
        Sorted list with highest-quality sources first.
    """
    return sorted(sources, key=score_author_source, reverse=True)


@dataclass(repr=False)
class SourcedAuthor(Author):
    """
    An author aggregated from multiple database sources.

    Combines author information from HAL, DBLP, and/or LDB into a single entity.
    The primary source (first in the sorted list) determines the author's key.

    Parameters
    ----------
    name: :class:`str`
        The author's name.
    sources: :class:`list`
        List of database-specific author objects (HALAuthor, DBLPAuthor, LDBAuthor).
    """

    sources: list = field(default_factory=list)

    @property
    def key(self):
        if self.sources:
            return self.sources[0].key
        else:
            return None

    @property
    def aliases(self):
        if self.sources:
            return clean_aliases(
                self.name, [n for a in self.sources for n in [a.name] + a.aliases]
            )
        else:
            return []

    @classmethod
    def from_sources(cls, sources):
        sources = sort_author_sources(sources)
        return cls(name=sources[0].name, sources=sources)

    def get_publications(self, clean=True, selector=None):
        if selector is None:
            selector = []
        if not isinstance(selector, list):
            selector = [selector]
        res = {
            p.key: p
            for a in self.sources
            for p in a.get_publications()
            if all(f(p) for f in selector)
        }
        if clean:
            regroup_authors({self.key: self}, res)
            return regroup_publications(res)
        else:
            return res


publication_score_rosetta = {
    "db_name": {"dblp": 1, "ldb": 1, "hal": 2},
    "venue": {"CoRR": -1, "unpublished": -2},
    "type": {"conference": 1, "journal": 2},
}


def score_publication_source(source):
    scores = [
        v.get(getattr(source, k, None), 0) for k, v in publication_score_rosetta.items()
    ]
    scores.append(source.year)
    return tuple(scores)


def sort_publication_sources(sources):
    return sorted(sources, key=score_publication_source, reverse=True)


@dataclass(repr=False)
class SourcedPublication(Publication):
    """
    A publication aggregated from multiple database sources.

    Combines publication entries from HAL, DBLP, and/or LDB that refer
    to the same paper. The primary source determines the publication's metadata.

    Parameters
    ----------
    title: :class:`str`
        Publication title.
    authors: :class:`list`
        List of author objects.
    venue: :class:`str`
        Publication venue.
    type: :class:`str`
        Publication type.
    year: :class:`int`
        Publication year.
    sources: :class:`list`
        List of database-specific publication objects.
    """

    sources: list = field(default_factory=list)

    @property
    def key(self):
        if self.sources:
            return self.sources[0].key
        else:
            return None

    @classmethod
    def from_sources(cls, sources):
        sources = sort_publication_sources(sources)
        main = sources[0]
        res = cls(
            **{
                k: getattr(main, k)
                for k in ["title", "authors", "venue", "type", "year"]
            },
            sources=sources,
        )
        return res


def regroup_authors(auth_dict, pub_dict):
    """
    Replace authors of publications with matching authors.
    Typical use: upgrade DB-specific authors to multisource authors.

    Replacement is in place.

    Parameters
    ----------
    auth_dict: :class:`dict`
        Authors to unify.
    pub_dict: :class:`dict`
        Publications to unify.

    Returns
    -------
    None
    """
    redirection = {
        k: a
        for a in auth_dict.values()
        for s in a.sources
        for k in [s.key, s.name, *s.aliases]
    }

    for pub in pub_dict.values():
        pub.authors = [
            redirection.get(a.key, redirection.get(a.name, a)) for a in pub.authors
        ]


def regroup_publications(pub_dict, threshold=85, length_impact=0.05, n_range=5):
    """
    Puts together copies of the same publication.

    Parameters
    ----------
    pub_dict: :class:`dict`
        Publications to unify.
    threshold: float
        Similarity parameter.
    length_impact: float
        Length impact parameter.

    Returns
    -------
    :class:`dict`
        Unified publications.
    """
    if len(pub_dict) == 0:
        return dict()
    pub_list = [p for p in pub_dict.values()]
    res = dict()
    vectorizer = CountVectorizer(n_range=n_range)
    x = vectorizer.fit_transform([unicodedata.normalize("NFKC", p.title) for p in pub_list])
    y = x.T.tocsr()
    jc_matrix = jit_square_factors(
        x.indices, x.indptr, y.indices, y.indptr, len(pub_list), length_impact
    )
    done = np.zeros(len(pub_list), dtype=bool)
    for i, paper in enumerate(pub_list):
        if done[i]:
            continue
        locs = np.where(jc_matrix[i, :] > threshold)[0]
        pub = SourcedPublication.from_sources([pub_list[i] for i in locs])
        res[pub.key] = pub
        done[locs] = True
    return res



from string import Template


# language=javascript
draw_script = """
import { DataSet, Network } from $vis_url;

const nodes = new DataSet($nodes);
const edges = new DataSet($edges);
const options = $options;
const container = document.getElementById('vis-$uid');
let hoveredEdgeId = null;
let hoveredNodeId = null;

// Get the group color and position of a node. Useful for gradient edges
function getNodeInfos(network, node) {
    if (node && !options.groups?.[node.group]?.hidden) {
        return [options.groups[node.group]?.color, network.getPositions([node.id])[node.id]]
    }
    return [false, false];
}


// main course
function draw_graph() {
    // No clean redraw so far, so we re-create everything :(
    // Set hidden groups according to legend, if any
    document.querySelectorAll('#legend-$uid .legend-checkbox').forEach(cb => {
    const group = cb.getAttribute('data-group');
    options.groups[group].hidden = !cb.checked;
        });


    // First compute the nodes to display.
    var visibleNodes = new DataSet(nodes.get({
      filter: node => !options.groups?.[node.group]?.hidden}));
    var visibleNodeIds = new Set(visibleNodes.map(node => node.id));
    // Reduce edges
    const visibleEdges = new DataSet(edges.get({
        filter: edge => visibleNodeIds.has(edge.from) && visibleNodeIds.has(edge.to)}));
    //  Optiional: remove comets
    if (!document.getElementById("comet-$uid")?.checked) {
        visibleNodeIds = new Set();
        visibleEdges.forEach(edge => {
            visibleNodeIds.add(edge.from);
            visibleNodeIds.add(edge.to);
        });
        visibleNodes = new DataSet(nodes.get({filter: node => visibleNodeIds.has(node.id)}));
    }


    // Set graph, nodes, and edges
    const network = new Network(container, {nodes: visibleNodes, edges: visibleEdges}, options);
    network.once("afterDrawing", function () {
      network.fit({ maxZoomLevel: 3 });
    });
    const netNodes = network.body.data.nodes;
    const netEdges = network.body.data.edges;

    // Gradient edges
    network.on("beforeDrawing", function(ctx) {
        const selectedEdgeIds = network.getSelectedEdges();
        netEdges.forEach(edge => {
            const [fromColor, fromPos] = getNodeInfos(network, netNodes.get(edge.from))
            const [toColor, toPos] = getNodeInfos(network, netNodes.get(edge.to))

        if (fromColor && toColor && (fromColor !== toColor) && fromPos && toPos ) {
            let width = edge.width || 2;
            if (selectedEdgeIds.includes(edge.id) || hoveredEdgeId === edge.id || hoveredNodeId === edge.from || hoveredNodeId === edge.to) width *= 1.8;

            // Gradient
            const grad = ctx.createLinearGradient(fromPos.x, fromPos.y, toPos.x, toPos.y);
            grad.addColorStop(0, fromColor);
            grad.addColorStop(1, toColor);

            // Draw line
            ctx.save();
            ctx.strokeStyle = grad;
            ctx.lineWidth = width;
            ctx.beginPath();
            ctx.moveTo(fromPos.x, fromPos.y);
            ctx.lineTo(toPos.x, toPos.y);
            ctx.stroke();
            ctx.restore();
        }
    });
});

    // Hover tooltip & record
    network.on("hoverEdge", params => {
        const edge = netEdges.get(params.edge);
        network.body.container.title = edge.hover || '';
        hoveredEdgeId = params.edge;
    });

    network.on("blurEdge", params => {
        network.body.container.title = '';
        hoveredEdgeId = null;
    });

    network.on("hoverNode", params => {
        const node = netNodes.get(params.node);
        netNodes.update({id: node.id, borderWidth: 10})
        network.body.container.title = node.hover || '';
        hoveredNodeId = params.node;
    });
    network.on("blurNode", params => {
        const node = netNodes.get(params.node);
        netNodes.update({id: node.id, borderWidth: 5})
        network.body.container.title = '';
        hoveredNodeId = null;
    });


    // Modal overlay
    const modal = document.getElementById('modal-$uid');
    const modalBody = document.getElementById('modal-body-$uid');
    const modalClose = document.getElementById('modal-close-$uid');
    network.on("click", function(params) {
      if (params.nodes.length === 1) {
        const node = netNodes.get(params.nodes[0]);
        modalBody.innerHTML = node.overlay || '';
        modal.style.display = "block";
      } else if (params.edges.length === 1) {
        const edge = netEdges.get(params.edges[0]);
        modalBody.innerHTML = edge.overlay || '';
        modal.style.display = "block";
      } else {
        modal.style.display = "none";
      }
    });
    modalClose.onclick = function() { modal.style.display = "none"; };
    window.onclick = function(event) {
      if (event.target == modal) { modal.style.display = "none"; }
    };
}

draw_graph();

"""

# language=javascript
redraw_script = """
document.getElementById('redraw-$uid').addEventListener('click', function(event) {
    event.preventDefault();  // Prevent page jump
    draw_graph();
});

"""

# language=javascript
fs_script = """
document.getElementById('fullscreen-$uid').addEventListener('click', function(event) {
    event.preventDefault();
    let elem = document.getElementById('box-$uid');
    if (!document.fullscreenElement) {
        // Request fullscreen mode
        if (elem.requestFullscreen) {
            elem.requestFullscreen();
        } else if (elem.webkitRequestFullscreen) { /* Safari */
            elem.webkitRequestFullscreen();
        } else if (elem.msRequestFullscreen) { /* IE11 */
            elem.msRequestFullscreen();
        }
    } else {
        // Exit fullscreen mode
        if (document.exitFullscreen) {
            document.exitFullscreen();
        } else if (document.webkitExitFullscreen) { /* Safari */
            document.webkitExitFullscreen();
        } else if (document.msExitFullscreen) { /* IE11 */
            document.msExitFullscreen();
        }
    }
});

"""

# language=javascript
legend_script = """
// Refresh when boxes are changed
document.querySelectorAll("#legend-$uid input").forEach(cb => {
  cb.addEventListener('change', draw_graph);
});


"""


default_script = Template(draw_script + redraw_script + fs_script + legend_script)



from domonic import tags
import distinctipy


def auto_groups(lab, groups=None, rng=None, pastel_factor=0.3):
    if groups is None:
        groups = ego_groups
    else:
        for k, v in ego_groups.items():
            if k not in groups:
                groups[k] = v
            else:
                groups[k] = {**v, **groups[k]}
    res = {
        group: groups.get(group, {"hidden": False})
        for i, group in enumerate(
            {
                a.metadata.group: None
                for a in lab.authors.values()
                if a and a.metadata.group
            }
        )
    }
    n_colors = len([None for g in res.values() if "color" not in g])
    colors = distinctipy.get_colors(n_colors, pastel_factor=pastel_factor, rng=rng)
    colors = [
        f"rgb({int(r * 255)},{int(g * 255)},{int(b * 255)})" for r, g, b in colors
    ]
    i = 0
    for group in res.values():
        if "color" not in group:
            group["color"] = colors[i]
            i += 1
    return res


ego_groups = {
    "star": {"display": "Star", "color": "rgb(210, 190, 70)", "hidden": False},
    "planet": {"display": "Planets", "color": "rgb(80, 120, 200)", "hidden": False},
    "moon": {"display": "Moons", "color": "rgb(140, 140, 140)", "hidden": False},
}


def make_legend(groups, uid):
    legend = tags.div(_id=f"legend-{uid}", _class="legend")
    if len(groups) > 1:
        for group_name, props in groups.items():
            color = props.get("color", "#cccccc")
            display_name = props.get("display", group_name)
            entry = tags.label(display_name, _class="legend-entry")
            entry.appendChild(
                tags.input(
                    **{
                        "type": "checkbox",
                        "class": "legend-checkbox",
                        "data-group": group_name,
                    },
                    checked=True,
                )
            )
            entry.appendChild(
                tags.span(
                    _style=f"background-color: {color}; width: 14px; height: 14px; display: inline-block; margin-right: 5px; vertical-align: middle;"
                )
            )
            legend.appendChild(entry)
    entry = tags.label("Show Comets", _class="comet-entry")
    entry.appendChild(tags.input(**{"type": "checkbox", "id": f"comet-{uid}"}))
    legend.appendChild(entry)
    return legend



from gismap.lab.labmap import LabMap
from gismap.lab.lab_author import LabAuthor


class EgoMap(LabMap):
    """
    Egocentric view of a researcher's collaboration network.

    Displays the *star* (central researcher), their *planets* (direct co-authors),
    and optionally *moons* (co-authors of co-authors).

    Parameters
    ----------
    star: :class:`str` or :class:`~gismap.lab.lab_author.LabAuthor`
        The central researcher. Can be a name string or LabAuthor object.
    *args
        Passed to :class:`~gismap.lab.labmap.LabMap`.
    **kwargs
        Passed to :class:`~gismap.lab.labmap.LabMap`.

    Examples
    --------

    >>> dang = EgoMap("The-Dang Huynh")
    >>> dang.build(target=20)
    >>> sorted(a.name for a in dang.authors.values() if len(a.name.split())<3)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    ['Bruno Kauffmann', 'Diego Perino', 'Dohy Hong', 'Fabien Mathieu', 'François Baccelli',...]
    """

    def __init__(self, star, *args, **kwargs):
        if isinstance(star, str):
            star = LabAuthor(star)
        star.metadata.position = (0, 0)
        star.metadata.group = "star"
        self.star = star
        super().__init__(*args, **kwargs)

    def _author_iterator(self):
        yield self.star

    def build(self, **kwargs):
        """
        Build the ego network by fetching publications and adding planets/moons.

        Parameters
        ----------
        target : :class:`int`, default=50
            Target number of authors in the final map.
        **kwargs
            Passed to :meth:`~gismap.lab.labmap.LabMap.expand`.

        Returns
        -------
        None
        """
        target = kwargs.pop("target", 50)
        self.update_authors(desc="Star metadata")
        self.update_publis(desc="Star publications")
        kwargs["target"] = target - len(self.authors)
        self.expand(group="planet", desc="Planets", **kwargs)
        kwargs["target"] = target - len(self.authors)
        if kwargs["target"] > 0:
            self.expand(group="moon", desc="Moons", **kwargs)



physics = {
    "solver": "forceAtlas2Based",
    "forceAtlas2Based": {
        "gravitationalConstant": -50,
        "centralGravity": 0.01,
        "springLength": 200,
        "springConstant": 0.08,
        "damping": 0.98,
        "avoidOverlap": 1,
    },
    "maxVelocity": 10,
    "minVelocity": 0.9,
    "stabilization": {
        "enabled": True,
        "iterations": 500,
        "updateInterval": 50,
        "onlyDynamicEdges": False,
        "fit": True,
    },
    "timestep": 0.25,
}

nodes = {
    "shape": "circle",
    "size": 20,
    "font": {"size": 16, "color": "#111"},
    "color": "rgb(59, 101, 178)",
    "borderWidth": 5,
}

edges = {
    "width": 2,
    # 'color': {'color': '#888', 'highlight': '#f5a25d'},
    "smooth": False,  # {"type": 'continuous'}
}

interaction = {"hover": True}



from bs4 import BeautifulSoup as Soup

from gismap.lab import LabAuthor
from gismap.lab.lab_author import AuthorMetadata
from gismap.lab.labmap import LabMap
from gismap.utils.requests import get


def lamsade_parse(div):
    """
    Parameters
    ----------
    div: :class:`~bs4.BeautifulSoup`
        Soup of the div of one researcher

    Returns
    -------
    :class:`tuple`
        name, image url (or None), webpage (or None)
    """
    img = div.img["src"] if div.img else None
    url = div.a["href"] if div.a else None
    name = div.h2.text.strip().title()
    name = " ".join(name.split(" ", 1)[::-1])
    return name, img, url


class Lamsade(LabMap):
    """
    Class for handling the Lamsade team (Dauphine).
    """

    name = "Lamsade"
    base_url = "https://www.lamsade.dauphine.fr/"
    directory = "fr/personnes/enseignants-chercheurs-et-chercheurs.html"

    def _author_iterator(self):
        soup = Soup(get(self.base_url + self.directory), features="lxml")
        for a in soup("div", class_="dauphinecv-item"):
            name, img, url = lamsade_parse(a)
            img = self.base_url + img if img else None
            url = self.base_url + url if url else None
            yield LabAuthor(
                name=name, metadata=AuthorMetadata(url=url, img=img, group=self.name)
            )



import re
import unicodedata
import base64
from IPython.display import display, HTML
import ipywidgets as widgets
from contextlib import contextmanager

from gismap.lab.egomap import EgoMap
from gismap.lab.labmap import ListMap


@contextmanager
def dummy_context():
    yield


def safe_filename(name):
    """
    Parameters
    ----------
    name: :class:`str`
        Pretty much anything.

    Returns
    -------
    :class:`str`
        GisMap filename.
    """
    normalized = unicodedata.normalize("NFKD", name)
    ascii_only = normalized.encode("ascii", "ignore").decode("ascii")
    ascii_only = ascii_only.replace(" ", "_")
    safe_str = re.sub(r"[^a-zA-Z0-9_]", "", ascii_only)
    return f"gismap-{safe_str[:60]}.html"


place_holder = "Diego Perino, The-Dang Huynh, François Durand (hal: fradurand, dblp: 38/11269), Rim Kaddah, Leonardo Linguaglossa, Céline Comte"


class GismapWidget:
    """
    A simple widget to test the production of LabMaps and EgoMaps.

    Examples
    --------

    This is a doctest example. Use a notebook to play with the widget.

    >>> gw = GismapWidget()  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    VBox(children=(HTML(value=''), Output(), HBox(children=(Textarea(value='', ...
    >>> gw.names.value = "Fabien Mathieu"
    >>> gw.dbs.value = "HAL"
    >>> gw.size.value = 3
    >>> gw.compute_function(gw.compute, show=False)
    >>> gw.save_link.value[:30]
    "<a href='data:text/html;base64"
    >>> gw.names.value = "Diego Perino, Laurent Viennot"
    >>> gw.compute_function(gw.compute, show=False)
    >>> gw.save_link.value[:30]
    "<a href='data:text/html;base64"
    """

    def __init__(self):
        self.names = widgets.Textarea(
            placeholder=place_holder,
            description="Name(s):",
            layout=widgets.Layout(width="50%", height="100px"),
        )
        self.dbs = widgets.RadioButtons(
            options=["HAL", "DBLP", "Both"],
            description="DB(s):",
            layout=widgets.Layout(width="80px", max_width="20%"),
        )
        self.size = widgets.IntSlider(
            value=10,
            min=1,
            max=150,
            step=1,
            description="Size",
            layout=widgets.Layout(width="250px"),
        )
        self.compute = widgets.Button(
            description="Map!", layout=widgets.Layout(width="120px", max_width="140px")
        )
        self._col = widgets.VBox(
            [self.size, self.compute],
            layout=widgets.Layout(
                align_items="center", max_width="27%", overflow="hidden"
            ),
        )
        self.save_link = widgets.HTML(value="")
        self.compute.on_click(self.compute_function)
        self.out = widgets.Output()
        self.widget = widgets.VBox(
            [self.save_link, self.out, widgets.HBox([self.names, self._col, self.dbs])]
        )
        display(self.widget)
        self.show = True

    def html(self):
        """
        Generate the HTML visualization based on widget inputs.

        Returns
        -------
        :class:`str`
            HTML content for the collaboration graph.
        """
        dbs = (
            "hal"
            if self.dbs.value == "HAL"
            else "dblp"
            if self.dbs.value == "DBLP"
            else ["hal", "dblp"]
        )
        name = self.names.value
        pattern = r",\s*(?![^()]*\))"
        names = [n.strip() for n in re.split(pattern, name)]
        self.save_link.value = ""
        ctx = self.out if self.show else dummy_context()
        if len(names) > 1:
            lab = ListMap(names, dbs=dbs, name="planet")
            if self.show:
                self.out.clear_output()
            with ctx:
                lab.update_authors()
                lab.update_publis()
                extra = self.size.value - len(lab.authors)
                if extra > 0:
                    lab.expand(target=extra)
        else:
            lab = EgoMap(names[0], dbs=dbs)
            if self.show:
                self.out.clear_output()
            with ctx:
                lab.build(target=self.size.value)
        return lab.html()

    def compute_function(self, b, show=True):
        """
        Handle compute button click and generate visualization.

        Parameters
        ----------
        b : :class:`ipywidgets.Button`
            The button widget that triggered this callback.
        show : :class:`bool`, default=True
            Whether to display the result in the widget.

        Returns
        -------
        None
        """
        self.show = show
        full = self.html()
        b64 = base64.b64encode(
            f"<html><body>{full}</body></html>".encode("utf8")
        ).decode("utf8")
        payload = f"data:text/html;base64,{b64}"
        savename = safe_filename(self.names.value)
        link_html = f"<a href='{payload}' download='{savename}'>Download the Map!</a>"
        self.save_link.value = link_html
        if show:
            self.out.clear_output()
            with self.out:
                display(HTML(full))



from bs4 import BeautifulSoup as Soup
from gismap.utils.requests import get
from gismap.lab.filters import re_filter
from gismap.lab import LabAuthor, Map


ghosts = [
    "Altman",
    "Lelarge",
    "Teixera",
    "Friedman",
    "Fdida",
    "Blaszczyszyn",
    "Jacquet",
    "Panafieu",
    "Bušić",
]
no_ghost = re_filter(ghosts)


class LINCS(Map):
    name = "LINCS"

    def _author_iterator(self):
        soup = Soup(get("https://www.lincs.fr/people/"), features="lxml")
        for entry in soup.main("div", class_="trombinoscope-row"):
            cols = entry("div")
            name = cols[1].text
            if not no_ghost(name):
                continue
            img = cols[0].img["src"]
            url = cols[-1].a
            if url:
                url = url.get("href")
            group = cols[2]("a")
            if group:
                group = group[-1].text
            else:
                group = "External"
            author = LabAuthor(name)
            author.metadata.img = img
            author.metadata.group = group
            author.metadata.url = url
            yield author



from bs4 import BeautifulSoup as Soup
from gismap.utils.common import get_classes
from gismap.lab import LabAuthor
from gismap.lab.lab_author import AuthorMetadata
from gismap.lab.labmap import LabMap
from gismap.utils.requests import get


def strip_li(li):
    """
    Parameters
    ----------
    li: :class:`~bs4.Tag`
        HTML list item

    Returns
    -------
    :class:`str`
        Searcher name
    """
    return li.text.split(',')[0].strip()


class CoTel(LabMap):
    """
    Abstract class for Algotel/Cores PC committee maps.

    Cf https://algotel.eu.org/
    """
    year = None
    @property
    def name(self):
        return f"algotel_cores_{self.year}"


_algotel_2026 = [
    "Achour Mostéfaoui", "Anaïs Durand", "Anne-Cécile Orgerie", "Antonella Del Pozzo",
    "Arnaud Casteigts", "Arnaud Labourel", "Cédric Bentz", "Colette Johnen",
    "Cristel Pelsser", "Emmanuelle Anceaume", "Fabien Mathieu", "Gewu Bu",
    "Hervé Rivano", "Hicham Khalifé", "Isabelle Guérin Lassous", "Jérémie Chalopin",
    "Jérémie Decouchant", "Jérémie Leguay", "Joanna Moulierac", "Josiane Kouam",
    "Lélia Blin", "Marc-Olivier Buob", "Nancy Perrot", "Pascal Berthou",
    "Pascal Felber", "Sara Tucci Piergiovanni", "Sébastien Tixeuil", "Stéphane Rovedakis",
    "Swan Dubois", "Vania Conan", "Xavier Défago",
]

_cores_2026 = [
    "André-Luc Beylot", "Anne Fladenmuller", "Bertrand Ducourthial", "Emmanuel Lavinal",
    "Francesco Bronzino", "Francois Lemercier", "Gérard Chalhoub", "Ghada Jaber",
    "Hind Castel", "Juan Fraire", "Julien Montavont", "Laurent Chasserat",
    "Lynda Zitoune", "Nathalie Mitton", "Oana Iova", "Pedro Velloso",
    "Rahim Kacimi", "Razvan Stanica", "Réjane Dalce", "Sahar Hoteit",
    "Sara Berri", "Tara Yahia", "Thai-Mai-Trang Nguyen", "Thibault Cholez",
    "Véronique Vèque", "Yassine Hadjadj-Aoul",
]


class AlgoRes2026(CoTel):
    """
    2026 edition. Website not yet public; PC hardcoded from internal source.
    """
    year = 2026
    def _author_iterator(self):
        for names, group in [(_algotel_2026, "algotel"), (_cores_2026, "cores")]:
            for searcher in names:
                yield LabAuthor(name=searcher, metadata=AuthorMetadata(group=group.title()))


class AlgoRes2025(CoTel):
    """
    https://algotel-cores25.sciencesconf.org/
    """
    year = 2025
    def _author_iterator(self):
        for i, group in [(13, "cores"), (9, "algotel")]:
            soup = Soup(get(f"https://algotel-cores25.sciencesconf.org/resource/page/id/{i}"), features="lxml")
            for searcher in [strip_li(li) for li in soup.ul("li")]:
                yield LabAuthor(name=searcher, metadata=AuthorMetadata(group=group.title()))


class AlgoRes2024(CoTel):
    """
    https://algotelcores2024.sciencesconf.org
    """
    year = 2024
    def _author_iterator(self):
        for group in ["cores", "algotel"]:
            soup = Soup(get(f"https://algotelcores2024.sciencesconf.org/data/{group}.html", encoding="utf-8"), features="lxml")
            for searcher in soup("ul", {"class": "person"}):
                yield LabAuthor(name=strip_li(searcher.li), metadata=AuthorMetadata(group=group.title()))


class AlgoRes2023(CoTel):
    """
    https://coresalgotel2023.i3s.univ-cotedazur.fr
    """
    year = 2023
    def _author_iterator(self):
        soup = Soup(get("https://coresalgotel2023.i3s.univ-cotedazur.fr/comites.html", encoding="utf-8"), features="lxml")
        for group, searchers in zip(["algotel", "cores"], [ul("li") for ul in soup.table("ul")]):
            for searcher in searchers:
                yield LabAuthor(name=strip_li(searcher), metadata=AuthorMetadata(group=group.title()))


class AlgoRes2022(CoTel):
    """
    https://sites.google.com/view/algotel-cores-2022
    """
    year = 2022
    def _author_iterator(self):
        for group in ["cores", "algotel"]:
            soup = Soup(get(f"https://sites.google.com/view/algotel-cores-2022/{group}-2022/comités"), features="lxml")
            for searcher in soup("ul")[-1]("li"):
                yield LabAuthor(name=strip_li(searcher), metadata=AuthorMetadata(group=group.title()))


class AlgoRes2021(CoTel):
    """
    https://apps.univ-lr.fr/cgi-bin/WebObjects/Colloque.woa/wa/menu?code=2721&idMenu=10922
    """
    year = 2021
    def _author_iterator(self):
        for i, group in [(10953, "cores"), (10947, "algotel")]:
            soup = Soup(get(f"https://apps.univ-lr.fr/cgi-bin/WebObjects/Colloque.woa/wa/menu?code=2721&idMenu={i}"), features="lxml")
            for searcher in soup("ul")[-1]("li"):
                yield LabAuthor(name=strip_li(searcher), metadata=AuthorMetadata(group=group.title()))


class AlgoRes2020(CoTel):
    """
    https://cores-algotel-2020.imag.fr/
    """
    year = 2020
    rosetta = {"Eric": "Éric Gourdin"}

    def _author_iterator(self):
        for i, group in [("index188e.html?page_id=73", "cores"), ("index4404.html?page_id=70", "algotel")]:
            soup = Soup(get(f"https://cores-algotel-2020.imag.fr/{i}", encoding="utf-8"), features="lxml")
            for searcher in soup.ul("li"):
                searcher = strip_li(searcher)
                searcher = self.rosetta.get(searcher, searcher)
                yield LabAuthor(name=searcher, metadata=AuthorMetadata(group=group.title()))


class AlgoRes2019(CoTel):
    """
    https://www.irit.fr/algotel2019/
    https://www.irit.fr/cores2019/index.html
    """
    year = 2019
    def _author_iterator(self):
        for url, group in [
            ("https://www.irit.fr/algotel2019/", "algotel"),
            ("https://www.irit.fr/cores2019/index.html", "cores"),
        ]:
            soup = Soup(get(url), features="lxml")
            for h3 in soup("h3"):
                if "programme" in h3.text.lower():
                    for p in h3.find_next_sibling("div")("p"):
                        yield LabAuthor(name=p.text.split(",")[0].strip(), metadata=AuthorMetadata(group=group.title()))
                    break


class AlgoRes2018(CoTel):
    """
    https://algotel2018.complexnetworks.fr/comite.html
    http://cores2018.complexnetworks.fr/comite.html
    """
    year = 2018
    rosetta = {"Gourdin": "Eric Gourdin", "Carneiro Viana": "Aline Carneiro Viana"}
    def _author_iterator(self):
        # Algotel: <p> tags after "Comité de programme" h2
        soup = Soup(get("https://algotel2018.complexnetworks.fr/comite.html", encoding="utf-8"), features="lxml")
        for h in soup("h2"):
            if "programme" in h.text.lower():
                for p in h.find_next_siblings("p"):
                    name = p.text.split(",")[0].strip()
                    name = self.rosetta.get(name, name)
                    yield LabAuthor(name=name, metadata=AuthorMetadata(group="Algotel"))
                break
        # Cores: <p> tags directly (single section page)
        soup = Soup(get("http://cores2018.complexnetworks.fr/comite.html", encoding="utf-8"), features="lxml")
        for p in soup("p"):
            name = p.text.split(",")[0].strip()
            if name:
                yield LabAuthor(name=name, metadata=AuthorMetadata(group="Cores"))


_algotel_2017 = [
    "Aline Carneiro Viana", "Stephane Devismes", "Karine Altisen", "Emmanuelle Anceaume",
    "Farid Benbadis", "Janna Burman", "Arnaud Casteigts", "Hakima Chaouchi",
    "David Coudert", "Sylvie Delaet", "Yoann Dieudonne", "Hugues Fauconnier",
    "Emmanuel Godard", "Vincent Gramoli", "Frederic Guinand", "Luis Henrique",
    "Sahar Hoteit", "Luigi Iannone", "Oana Iova", "Hicham Khalife",
    "Ralf Klasing", "Pascal Lafourcade", "Patrick Maille", "Nathalie Mitton",
    "Achour Mostefaoui", "Anelise Munaretto", "Franck Petit", "Stephane Rovedakis",
    "Damien Saucez", "Thomas Silverston", "Fabrice Theoleyre", "Gilles Tredan",
    "Artur Ziviani",
]

_cores_2017 = [
    "Franck Rousseau", "Fabrice Valois", "Dominique Barthel", "Andre-Luc Beylot",
    "Yann Busnel", "Isabelle Chrisment", "Vania Conan", "Marcelo Dias de Amorim",
    "Bertrand Ducourthial", "Andrzej Duda", "Eric Fleury", "Jean-Loup Guillaume",
    "Fabrice Guillemin", "Alexandre Guitton", "Katia Jaffres-Runser", "Houda Labiod",
    "Xavier Lagrange", "Pascale Minet", "Nathalie Mitton", "Thomas Noel",
    "Georgios Papadopoulos", "Cristel Pellser", "Bruno Quoitin", "Razvan Stanica",
    "Thierry Turletti", "Pascal Thubert", "Guillaume Urvoy-Keller", "Veronique Veque",
]


class AlgoRes2017(CoTel):
    """
    https://web.archive.org/web/20170719001105/http://algotel2017.ensai.fr/
    https://web.archive.org/web/20171203000444/http://cores2017.ensai.fr/
    """
    year = 2017
    def _author_iterator(self):
        for names, group in [(_algotel_2017, "algotel"), (_cores_2017, "cores")]:
            for searcher in names:
                yield LabAuthor(name=searcher, metadata=AuthorMetadata(group=group.title()))


class AlgoRes2016(CoTel):
    """
    https://algotel2016.labri.fr/index.php?committee
    https://algotel2016.labri.fr/index.php?cores
    """
    year = 2016
    rosetta = {"Laurent Viennot LIAFA": "Laurent Viennot"}
    def _author_iterator(self):
        # Algotel: committee names in the largest <ul> with 30 <li>
        soup = Soup(get("https://algotel2016.labri.fr/index.php?committee"), features="lxml")
        ul = max(soup("ul"), key=lambda u: len(u("li")))
        for li in ul("li"):
            name = strip_li(li)
            name = self.rosetta.get(name, name)
            yield LabAuthor(name=name, metadata=AuthorMetadata(group="Algotel"))
        # Cores: <p> tags after "Comité scientifique" h3
        soup = Soup(get("https://algotel2016.labri.fr/index.php?cores"), features="lxml")
        for h in soup("h3"):
            if "scientifique" in h.text.lower():
                for p in h.find_next_siblings("p"):
                    name = p.text.split(",")[0].strip()
                    if name:
                        yield LabAuthor(name=name, metadata=AuthorMetadata(group="Cores"))
                break


cotels = get_classes(CoTel, key="year")



from dataclasses import dataclass

HIDDEN_KEYS = {"sources", "aliases", "abstract", "metadata"}


class LazyRepr:
    """
    MixIn that provides a clean repr for dataclasses.

    Hides empty fields and fields in HIDDEN_KEYS from the repr string.
    Private attributes (starting with '_') are also hidden.
    """

    def __repr__(self):
        kws = [
            f"{key}={value!r}"
            for key, value in self.__dict__.items()
            if value and key not in HIDDEN_KEYS and not key.startswith("_")
        ]
        return f"{type(self).__name__}({', '.join(kws)})"


def unlist(x):
    """
    Parameters
    ----------
    x: :class:`str` or :class:`list` or :class:`int`
        Something.

    Returns
    -------
    x: :class:`str` or :class:`int`
        If it's a list, make it flat.
    """
    return x[0] if (isinstance(x, list) and x) else x


def get_classes(root, key="name", recurse=False):
    """
    Parameters
    ----------
    root: :class:`class`
        Starting class (can be abstract).
    key: :class:`str`, default='name'
        Attribute to look-up
    recurse: bool, default=False
        Recursively traverse subclasses.

    Returns
    -------
    :class:`dict`
        Dictionaries of all subclasses that have a key attribute (as in class attribute `key`).

    Examples
    --------

    >>> from gismap.sources.models import DB
    >>> subclasses = get_classes(DB, key='db_name')
    >>> dict(sorted(subclasses.items())) # doctest: +NORMALIZE_WHITESPACE
    {'dblp': <class 'gismap.sources.dblp.DBLP'>,
    'hal': <class 'gismap.sources.hal.HAL'>,
    'ldb': <class 'gismap.sources.ldb.LDB'>}
    """
    result = {
        getattr(c, key): c for c in root.__subclasses__() if getattr(c, key, None)
    }
    if recurse:
        for c in root.__subclasses__():
            result.update(get_classes(c, key=key, recurse=True))
    return result


def list_of_objects(clss, dico, default=None):
    """
    Versatile way to enter a list of objects referenced by a dico.

    Parameters
    ----------
    clss: :class:`object`
        Object or reference to an object or list of objects / references to objects.
    dico: :class:`dict`
        Dictionary of references to objects.
    default: :class:`list`, optional
        Default list to return if `clss` is None.

    Returns
    -------
    :class:`list`
        Proper list of objects.

    Examples
    ________

    >>> from gismap.sources.models import DB
    >>> subclasses = get_classes(DB, key='db_name')
    >>> from gismap import HAL, DBLP, LDB
    >>> list_of_objects([HAL, 'ldb'], subclasses)
    [<class 'gismap.sources.hal.HAL'>, <class 'gismap.sources.ldb.LDB'>]
    >>> list_of_objects(None, subclasses, [DBLP])
    [<class 'gismap.sources.dblp.DBLP'>]
    >>> list_of_objects(LDB, subclasses)
    [<class 'gismap.sources.ldb.LDB'>]
    >>> list_of_objects('hal', subclasses)
    [<class 'gismap.sources.hal.HAL'>]
    """
    if default is None:
        default = []
    if clss is None:
        return list_of_objects(clss=default, dico=dico)
    elif isinstance(clss, str):
        return [dico[clss]]
    elif isinstance(clss, list):
        return [cls for lcls in clss for cls in list_of_objects(lcls, dico, default)]
    else:
        return [clss]


@dataclass(repr=False)
class Data(LazyRepr):
    """
    Easy-going converter of dict to dataclass. Useful when you want to use attribute access
    and do not care about giving a full description.

    Examples
    --------

    >>> data = Data({
    ... 'name': 'Alice',
    ... 'age': 30,
    ... 'address': {'street': '123 Main', 'city': 'Paris'},
    ... 'hobbies': [{'name': 'jazz', 'level': 5}, {'name': 'code'}]})
    >>> data # doctest: +NORMALIZE_WHITESPACE
    Data(name='Alice', age=30, address=Data(street='123 Main', city='Paris'),
    hobbies=[Data(name='jazz', level=5), Data(name='code')])
    >>> data.hobbies[0].name
    'jazz'
    >>> data.todict()  # doctest: +NORMALIZE_WHITESPACE
    {'name': 'Alice', 'age': 30, 'address': {'street': '123 Main', 'city': 'Paris'},
    'hobbies': [{'name': 'jazz', 'level': 5}, {'name': 'code'}]}
    """

    def __init__(self, data):
        self._wrap(data)

    def _wrap(self, d):
        for key, value in d.items():
            wrapped = self._wrap_item(value)
            setattr(self, key, wrapped)

    def _wrap_item(self, item):
        if isinstance(item, dict):
            return Data(item)
        elif isinstance(item, list):
            return [self._wrap_item(subitem) for subitem in item]
        return item

    def todict(self):
        res = {}
        for key in self.__dict__:
            value = getattr(self, key)
            if hasattr(value, "todict"):
                res[key] = value.todict()
            elif isinstance(value, list):
                res[key] = [v.todict() if hasattr(v, "todict") else v for v in value]
            else:
                res[key] = value
        return res



import requests
from bs4 import BeautifulSoup as Soup
from gismap.lab.labmap import LabMap
from gismap.lab.lab_author import AuthorMetadata, LabAuthor
from gismap.utils.requests import get


class CedricMap(LabMap):
    """
    Class for handling a CNAM Cedric team from its name.
    Default to `roc` team.
    """

    name = "roc"
    base_url = "https://cedric.cnam.fr"

    def _author_iterator(self):
        soup = Soup(get(f"{self.base_url}/equipes/{self.name}/"), features="lxml")
        searchers = [
            li.a
            for ul in soup.find("div", {"id": "annuaire"})("ul")[:3]
            for li in ul("li")
        ]
        done = set()
        for searcher in searchers:
            name = searcher.text.split("(")[0].strip()
            if name in done:
                continue
            url = f"{self.base_url}{searcher['href']}"
            sousoup = Soup(get(url), features="lxml")
            img = sousoup.find("img", {"class": "photo"})["src"]
            response = requests.head(img, allow_redirects=True)
            if int(response.headers.get("Content-Length")) < 3000:
                img = None
            done.add(name)
            yield LabAuthor(
                name=name,
                metadata=AuthorMetadata(url=url, img=img, group=self.name.upper()),
            )


class CedricFull(LabMap):
    """
    Class for handling all CNAM Cedric teams using `https://cedric.cnam.fr/equipes` to get team names.
    """

    name = "Cedric"

    def _author_iterator(self):
        base = "https://cedric.cnam.fr/equipes/"
        soup = Soup(get(base), features="lxml")
        teams = {
            a["href"].split("/")[-2]
            for a in soup("a")
            if base in a.get("href", "") and len(a["href"]) > len(base)
        }
        for team in teams:
            for author in CedricMap(name=team)._author_iterator():
                yield author
